=head1 NAME

Slapd::Backend::Cyrus - base class of slapd-cyrus

=head1 DESCRIPTION

Synchronization between LDAP and Cyrus.

=cut

package Slapd::Backend::Cyrus;

use strict;
use warnings;
use base 'Slapd::Backend::Proxy';

use Slapd::Backend::Cyrus::Cyradm;
use Net::LDAP::Constant qw( LDAP_SUCCESS LDAP_OPERATIONS_ERROR LDAP_ALREADY_EXISTS LDAP_NO_SUCH_OBJECT );

our $VERSION = '0.01';

__PACKAGE__->mk_accessors(qw/ cyradm /);

=head1 PROTECTED METHODS

=over

=item B<_init>()

Purpose:
  * Initialization

Arguments:
  * None

Returns:
  * Nonzero if initialization failed

=cut

sub _init
{
	my $self = shift;

	$self->cyradm = new Slapd::Backend::Cyrus::Cyradm( $self->{cyrus_server} )
		or die "Can't connect to server $self->{cyrus_server}\n";
	$self->cyradm->authenticate(
		-user     => $self->{cyrus_user},
		-password => $self->{cyrus_password},
	) or die $self->cyradm->error();

	return LDAP_SUCCESS;
}

=item B<_config>( $name, @values )

Purpose:
  * Config

Arguments:
  * Scalar - name
  * List - values

Returns:
  * Scalar - result code

=cut

sub _config
{
	my $self = shift;
	my( $name, @values ) = @_;
	if( $name eq 'cyrusServer' )
	{
		$self->{cyrus_server} = shift @values;
	}
	elsif( $name eq 'cyrusAuth' )
	{
		$self->{cyrus_user} = shift @values;
		$self->{cyrus_password} = shift @values;
	}
	elsif( $name eq 'cyrusHierSep' )
	{
		$self->{sep} = shift @values;
	}
	else
	{
		return LDAP_OPERATIONS_ERROR;
	}
	return LDAP_SUCCESS;
}

=item B<_cyradm_error>()

Purpose:
  * Translate cyrus error into ldap

Arguments:
  * None

Returns:
  * Scalar - ldap error code

=cut

sub _cyradm_error
{
	my( $self ) = @_;
	my $error = $self->cyradm->error();
	return LDAP_SUCCESS unless $error;
	my $err = $error;
	$err =~ s/^.*: //;
	my %err_map = (
		'Mailbox already exists' => LDAP_ALREADY_EXISTS,
		'Mailbox does not exist' => LDAP_NO_SUCH_OBJECT,
	);
	if( $err_map{$err} )
	{
		return $err_map{$err};
	}
	else
	{
		warn $error;
		return LDAP_OPERATIONS_ERROR;
	}
}

=back

=cut

1;

=head1 SEE ALSO

L<Slapd::Backend::Proxy>, L<slapd-perl>

=head1 AUTHOR

Aleksey Mashanov <aleksey.mashanov@gmail.com>

=head1 LICENSE

This library is free software, you can redistribute it and/or modify it under
the same terms as Perl itself.

=cut
